
	.text

	/*
	 * Transition from Real Mode to Protected Mode as outlined in the
	 * Intel(R) 64 and IA-32 Architectures Software Developer's Manual
	 * vol. 3A, chapter 9.9.1.  Lines prefixed with '>' are direct quotes.
	 * Annotations within those lines are enclosed in [square brackets].
	 *
	 * ATTENTION: This messes with the stack!  Before calling this function,
	 * you MUST ensure that %sp is aligned to 4 bytes.  The stack pointer
	 * will be automatically translated to its linear address equivalent
	 * (i.e. %esp = %ss * 0x10 + %sp), however, it DOES NOT touch %ebp.
	 *
	 * ATTENTION: You MUST call this from real mode (.code16), and any code
	 * following the call MUST be for 32-bit protected mode (.code32).
	 *
	 * ATTENTION: This trashes %eax, %edx, eflags, and ALL segment registers.
	 * Furthermore, it clears the IF flag (i.e. it disables interrupts).
	 *
	 > Intel 64 and IA-32 processors have slightly different requirements
	 > for switching to protected mode.  To ensure upwards and downwards
	 > code compatibility with Intel 64 and IA-32 processors, we recommend
	 > that you follow these steps:
	 */
GLOBL real_to_prot
	.code16

	/* okay we have to be a little careful with this one */

	/*
	 > 1. Disable interrupts.  A CLI instruction disables maskable hardware
	 >    interrupts.  NMI interrupts can be disabled with external circuitry.
	 *
	 * So we mask the Non-Maskable Interrupt.  Truly an x86 moment.
	 */
	cli
	in	$0x70, %al
	or	$0x80, %al
	out	%al, $0x70
	in	$0x71, %al

	/* Grab the return address and store the old %ss */
	xor	%edx, %edx
	pop	%dx		/* %dx = return address */
	mov	%ss, %ax
	mov	%ax, real_ss

	/*
	 > 2. Execute the LGDT instruction to load the GDTR register
	 >    with the base address of the GDT.
	 */
	lgdt	gdtr

	/*
	 > 3. Execute a MOV CR0 instruction that sets the PE flag
	 >    (and optionally the PG flag) in control register CR0.
	 *
	 * We don't use paging but we DO want to set the Protection Enable bit.
	 */
	mov	%cr0, %eax
	or	$1, %al
	mov	%eax, %cr0

	/*
	 > 4. Immediately following the MOV CR0 instruction, execute a far
	 >    JMP or far CALL instruction.  (This operation is typically a far
	 >    jump or call to the next instruction in the instruction stram.)
	 */
	ljmpl	$0x08, $1f
1:	.code32

	/* recompute the (now linear) stack address */
	xor	%eax, %eax
	mov	%ss, %ax
	shl	$4, %eax
	add	%eax, %esp

	/*
	 > 5. The JMP or CALL instruction immediately after the MOV CR0
	 >    instruction changes the flow of execution and serializes
	 >    the processor.
	 *
	 * Ok.
	 *
	 > 6. If paging is enabled [...]
	 *
	 * It is not.
	 *
	 > 7. If a local descriptor table is going to be used [...]
	 *
	 * No.
	 *
	 > 8. Execute the LTR instruction [...]
	 *
	 * We don't do tasks here
	 *
	 > 9. After entering protected mode, the segment registers continue
	 >    to hold the contents they had in real-address mode.
	 >    The JMP or CALL instruction in step 4 resets the CS register.
	 >    Perform one of the folowing operations to update the contents
	 >    of the remaining segment registers.
	 >
	 >    - Reload segment registers DS, SS, ES, FS, and GS.
	 >      If the ES, FS, and/or GS registers are not going to be used,
	 >      load them with a null selector.
	 >    - [...]
	 */

	mov	$0x10, %ax
	mov	%ax, %ds
	mov	%ax, %ss
	mov	%ax, %es
	xor	%eax, %eax
	mov	%ax, %fs
	mov	%ax, %gs

	/*
	 > 10. Execute the LIDT instruction to load the IDTR register
	 >     with the address and limit of the protected-mode IDT.
	 *
	 * Ok but before doing that, we need to stash away the old IDTR
	 * so we can return to Real Mode and make BIOS calls later.
	 */
	sidt	real_idtr
	lidt	prot_idtr

	/*
	 > 11. Execute the STI instruction to enable maskable hardware
	 >     interrupts and perform the necessary hardware operation
	 >     to enable NMI interrupts.
	 *
	 * We don't really do interrupts in stage1, but we can enable NMI.
	 */
	in	$0x70, %al
	and	$0x7f, %al
	out	%al, $0x70
	in	$0x71, %al

	jmp	*%edx
END real_to_prot

	/*
	 * Do the exact opposite of real_to_prot, i.e. transition from
	 * Protected Mode to Real Mode (in order to make BIOS calls).
	 * Again, this is quoted from the Intel SDM vol. 3A, but this time
	 * from chapter 9.9.2.
	 *
	 * ATTENTION: This messes with the stack!  Every call to prot_to_real
	 * MUST be followed by a call to real_to_prot, and that call MUST
	 * happen with the same stack pointer value.
	 *
	 > The processor switches from protected mode back to real-address
	 > mode if software clears the PE bit in the CR0 register with a
	 > MOV CR0 instruction.  A procedure that re-enters real-address mode
	 > should perform the following steps:
	 */
GLOBL prot_to_real
	.code32

	/*
	 > 1. Disable interrupts.  [...]
	 */
	cli
	in	$0x70, %al
	or	$0x80, %al
	out	%al, $0x70
	in	$0x71, %al

	/* fuck with the stack */
	pop	%edx		/* %edx = return address */
	xor	%eax, %eax
	mov	real_ss, %ax
	shl	$4, %eax
	sub	%eax, %esp

	/*
	 > 2. If paging is enabled, [...]
	 *
	 * It is not.
	 *
	 > 3. Transfer program control to a readable segment that has a limit
	 >    of 64 KBytes (FFFFH).  This operation loads the CS register with
	 >    the segment limit required in real-address mode.
	 */
	ljmp	$0x18, $1f
1:	.code16

	/*
	 > 4. Load segment registers SS, DS, ES, FS, and GS with a selector
	 >    for a descriptor containing the following values, which are
	 >    appropriate for real-address mode:
	 >
	 >    - [...]
	 *
	 * See offset 0x20 in _gdt.
	 */
	mov	$0x20, %ax
	mov	%ax, %ds
	mov	%ax, %ss
	mov	%ax, %es
	mov	%ax, %fs
	mov	%ax, %gs

	/*
	 > 5. Execute the LIDT instruction to point to a real-address mode interrupt
	 >    table that is within the 1-MByte real-address mode address range.
	 */
	sidt	prot_idtr
	lidt	real_idtr

	/*
	 > 6. Clear the PE flag in the CR0 register to switch to real-address mode.
	 */
	mov	%cr0, %eax
	and	$0xfe, %al
	mov	%eax, %cr0

	/*
	 > 7. Execute a far JMP instruction to jump to a real-address mode
	 >    program.  This operation flushes the instruction queue and loads
	 >    the appropriate base-address value in the CS register.
	 */
	ljmp	$0x00, $1f

	/*
	 > 8. Load the SS, DS, ES, FS, and GS registers as needed by the
	 >    real-address mode code.  If any of the registers are not going
	 >    to be used in real-address mode, write 0s to them.
	 */
1:	mov	real_ss, %ax
	mov	%ax, %ss
	xor	%ax, %ax
	mov	%ax, %ds
	mov	%ax, %es
	mov	%ax, %fs
	mov	%ax, %gs

	/*
	 > 9. Execute the STI instruction to enable maskable hardware
	 >    interrupts and perform the necessary hardware operation
	 >    to enable NMI interrupts.
	 */
	in	$0x70, %al
	and	$0x7f, %al
	out	%al, $0x70
	in	$0x71, %al

	jmp	*%dx
END prot_to_real
	.code32

	.data

	.align 32
LOCAL gdtr
	.word gdt_end - gdt - 1
	.long gdt
END gdtr

	/*
	 * This figure is based on the Intel SDM vol. 3A, fig. 5-1.
	 * Reordered so that the structure matches the definition below, i.e.
	 * little endian byte order (bits WITHIN the bytes are big endian).
	 *
	 *    |       0               1       |       2               3       |
	 *    +-------------------------------+-------------------------------+
	 *    |         Segment Limit         |         Base Address          |
	 * +0 |                               |                               |
	 *    |             00:15             |             00:15             |
	 *    +---------------+-+---+-+-+-+-+-+-+-+-+-+-------+---------------+
	 *    |     Base      | | D |1|0|E|W|A| |B| |A| Limit |     Base      | < data
	 * +4 |               |P| P |  Type   |G|/|0|V|       |               |
	 *    |     16:23     | | L |1|1|C|R|A| |D| |L| 16:19 |     24:31     | < code
	 *    +---------------+-+---+-+-+-+-+-+-+-+-+-+-------+---------------+
	 *    |       4       |       5       |       6       |       7       |
	 *
	 * A    Accessed                            E  Expansion Direction (0=up)
	 * AVL  Available to System Programmers     G  Granularity (0=1B, 1=4K)
	 * B    Big (32-bit code)                   R  Readable
	 * C    Conforming                          W  Writable
	 * D    Default                             P  Present
	 * DPL  Descriptor Privilege Level
	 *
	 * The GDT is a Very Normal data structure.
	 */
	.align 32
LOCAL gdt
/* 0x00 null descriptor */
	.word              0x0000,                         0x0000
	.byte       0x00,           0x00,           0x00,           0x00

/* 0x08 32-bit code */
	/*    |        Limit = 0xfffff        |       Base = 0x00000000       | */
	.word              0xffff,                         0x0000
	/*    |               |P 0 0 1|1   R  |G D 0 /|  0xf  |               | */
	.byte       0x00,           0x9a,           0xcf,           0x00

/* 0x10 32-bit data */
	/*    |        Limit = 0xfffff        |       Base = 0x00000000       | */
	.word              0xffff,                         0x0000
	/*    |               |P 0 0 1|0   W  |G B 0 /|  0xf  |               | */
	.byte       0x00,           0x92,           0xcf,           0x00

/* 0x18 16-bit code */
	/*    |        Limit = 0x0ffff        |       Base = 0x00000000       | */
	.word              0xffff,                         0x0000
	/*    |               |P 0 0 1|1 C R  |    0 /|  0x0  |               | */
	.byte       0x00,           0x9e,           0x00,           0x00

/* 0x20 16-bit data */
	/*    |        Limit = 0x0ffff        |       Base = 0x00000000       | */
	.word              0xffff,                         0x0000
	/*    |               |P 0 0 1|0   W  |    0 /|  0x0  |               | */
	.byte       0x00,           0x92,           0x00,           0x00
gdt_end:
END gdt

	.section .bss

/* dummy IDT Register value (0 entries) */
LOCAL prot_idtr
	.word 0
	.long 0
END prot_idtr

/* this is where we store the Real Mode IDT for later BIOS calls */
LOCAL real_idtr
	.word 0
	.long 0
END real_idtr

LOCAL real_ss
	.word 0
END real_ss
